/*
Copyright (C) 2015 Alexey Karyakin <alexey.karyakin@uwaterloo.ca>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "LinearAllocator.h"
#include "PageStore.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

enum { PAGE_BITS_1 = 21, PAGE_BITS_2 = 30 };

template <unsigned BITS>
LinearAllocator<BITS>::LinearAllocator(unsigned ring_bits, PageStore *page_store):
	m_page_store(page_store),
	m_ring_mask((size_t(1) << ring_bits) - 1u),
	m_head(0),
	m_blk_head(0),
	m_blk_tail(0),
	m_overhead(0),
	m_stat()
{
	assert(ring_bits > BITS);

	size_t limit = size_t(1) << ring_bits;
	size_t page_sz = ring_bits >= PAGE_BITS_2 ? (size_t(1) << PAGE_BITS_2)
			: ring_bits >= PAGE_BITS_1 ? (size_t(1) << PAGE_BITS_1)
			: 0;

	char * region = (char *)mmap(NULL, limit + page_sz, PROT_NONE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_NORESERVE, -1, 0);
	if (region == MAP_FAILED)
	{
		fprintf(stderr, "mmap (initial region) returned %p, errno %d\n", m_region, errno);
		abort();
	}

	m_base = page_sz != 0 ? (char *)((size_t(region) + page_sz - 1) & ~(page_sz - 1)) : region;
}

template <unsigned BITS>
LinearAllocator<BITS>::~LinearAllocator()
{
}

template <unsigned BITS>
typename LinearAllocator<BITS>::offset_t LinearAllocator<BITS>::allocate_slow(size_t size)
{
	assert(m_head + size > m_blk_head);
	assert(m_blk_head <= m_ring_mask + 1);
	assert((m_blk_head & BLOCK_MASK) == 0);
	if (m_blk_head == m_ring_mask + 1)
	{
		assert(m_overhead == 0);
		m_overhead = m_ring_mask + 1 - m_head;
		m_blk_head = 0;
		m_head = 0;
	}
	offset_t blk_next = m_blk_head + BLOCK_SZ;
	if ((blk_next & m_ring_mask) == m_blk_tail)
	{
		fprintf(stderr, "CommitQueue exhausted\n");
		abort();
	}
	void *reused = m_page_store ? m_page_store->get() : NULL;
	if (reused)
	{
		if (reused != m_base + blk_next)
		{
			void *p = mremap(reused, BLOCK_SZ, BLOCK_SZ, MREMAP_FIXED | MREMAP_MAYMOVE, m_base + m_blk_head);
			if (p != m_base + m_blk_head)
			{
				fprintf(stderr, "mremap returned %p, errno %d\n", p, errno);
				abort();
			}
			p = mmap(reused, BLOCK_SZ, PROT_NONE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_NORESERVE | MAP_FIXED, -1, 0);
			if (p != reused)
			{
				fprintf(stderr, "mmap (remapped region) returned %p, errno %d\n", p, errno);
				abort();
			}
		}
		m_stat.reused_blk++;
	}
	else
	{
		if (mprotect(m_base + m_blk_head, BLOCK_SZ, PROT_READ | PROT_WRITE) != 0)
		{
			fprintf(stderr, "mprotect (expand) failed, errno %d\n", errno);
			abort();
		}
		m_stat.mapped_blk += 1;
	}
	if (BITS >= PAGE_BITS_1)
		madvise(m_base + m_blk_head, BLOCK_SZ, m_blk_head != 0 ? MADV_HUGEPAGE : MADV_NOHUGEPAGE);
	// cause a page fault for accurate time measurements
	*(volatile char *)(m_base + m_blk_head) = '\0';
	m_blk_head = blk_next;
	auto p = m_head;
	m_head += size;
	return p;
}

template <unsigned BITS>
void LinearAllocator<BITS>::collect_blocks(offset_t blk_ofs)
{
	assert((blk_ofs & BLOCK_MASK) == 0);
	assert(blk_ofs != m_blk_tail);
	assert(blk_ofs < m_ring_mask + 1);

	while (m_blk_tail != blk_ofs)
	{
		void *p = m_base + m_blk_tail;
		if (m_page_store)
		{
			p = m_page_store->put(p);
			m_stat.returned_blk++;
		}
		if (p)
		{
			if (BITS >= PAGE_BITS_1 && madvise(p, BLOCK_SZ, MADV_NOHUGEPAGE)!= 0)
				fprintf(stderr, "madvise (nohugepage) failed, errno %d\n", errno);
			if (madvise(p, BLOCK_SZ, MADV_DONTNEED) != 0)
				fprintf(stderr, "madvise (free) failed, errno %d\n", errno);
			m_stat.unmapped_blk += 1;
		}
		m_blk_tail += BLOCK_SZ;
		m_blk_tail &= m_ring_mask;
		m_stat.collected_blk++;
		if (m_blk_tail == 0)
			m_overhead = 0;
	}
}

template class LinearAllocator<21>;
template class LinearAllocator<20>;
template class LinearAllocator<14>;
