/*
Copyright (C) 2015 Alexey Karyakin <alexey.karyakin@uwaterloo.ca>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef LinearAllocator_H
#define LinearAllocator_H

#include <stdlib.h>
#include <stdint.h> // stat

#include "likely.h"
#include <assert.h>

class PageStore;

template <unsigned BITS>
class LinearAllocator
{
public:

	struct Stat
	{
		uint64_t mapped_blk;
		uint64_t reused_blk;
		uint64_t unmapped_blk;
		uint64_t returned_blk;
		uint64_t collected_blk;
		uint64_t kernel_clk;
	};

	typedef uint64_t offset_t;

	enum : offset_t
	{
		BLOCK_SZ = offset_t(1) << BITS,
		BLOCK_MASK = BLOCK_SZ - 1
	};

	LinearAllocator(unsigned space_bits, PageStore *page_store);
	~LinearAllocator();

	offset_t allocate(size_t size)
	{
		assert(size <= BLOCK_SZ);
		assert((size & 7) == 0);
		assert(size != 0);
		offset_t p, next = m_head + size;
		if (likely(next <= m_blk_head))
		{
			p = m_head;
			m_head = next;
		}
		else
			p = allocate_slow(size);
		return p;
	}
	void collect(offset_t ofs) // NOTE: intended to be rarely called
	{
//		assert((ofs & 7) == 0);
		ofs &= ~BLOCK_MASK;
		if (unlikely(ofs != m_blk_tail))
			collect_blocks(ofs);
	}

	offset_t head() const { return m_head; }
	offset_t block_tail() const { return m_blk_tail; }

	void *resolve(offset_t ofs) const
	{
		assert(valid(ofs));
		return m_base + ofs;
	}
	offset_t offset(void *p) const
	{
		offset_t ofs = (char *) p - m_base;
		assert(valid(ofs));
		return ofs;
	}
	size_t size_gross() const // NOTE: including tail overhead and unfreed space in the first block
	{
		return (m_head - m_blk_tail) & m_ring_mask;
	}
	size_t diff_gross(offset_t a, offset_t b) const // NOTE: includes overhead
	{
		assert(valid(a));
		assert(valid(b));
		return (a - b) & m_ring_mask;
	}
	offset_t move(offset_t ofs, size_t el_sz, unsigned n = 1) const
	{
		assert(valid(ofs));
		offset_t total = el_sz * n;
		auto next = ofs + total;
		if (unlikely(next > m_ring_mask)) // NOTE: strictly greater than
		{
			offset_t available = m_ring_mask - ofs + 1u;
			next = total - available + available % el_sz;
		}
		return next;
	}
	offset_t move_back(offset_t ofs, size_t el_sz, unsigned n = 1) const
	{
		assert(valid(ofs));
		offset_t total = el_sz * n;
		if (unlikely(total > ofs))
			ofs += m_ring_mask + 1 - m_overhead - total;
		return ofs - total;
	}
	bool valid(offset_t offset) const // NOTE: weak check
	{
		return ((offset - m_blk_tail) & m_ring_mask) <= size_gross();
	}

	offset_t ring_mask() const { return m_ring_mask; }
	const Stat &stat() const { return m_stat; }

private:

	offset_t allocate_slow(size_t size);
	void collect_blocks(offset_t blk_ofs);

	PageStore *m_page_store;
	offset_t m_ring_mask;
	char *m_region;
	char *m_base;
	offset_t m_head;
	offset_t m_blk_head;
	offset_t m_blk_tail;
	offset_t m_overhead;
	Stat m_stat;

	friend class LinearAllocatorTest;
	friend class LinearAllocatorTestSmall;
};

#endif
