/*
Copyright (C) 2015 Alexey Karyakin <alexey.karyakin@uwaterloo.ca>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <map>
#include <vector>
#include <queue>
#include <cassert>
#include <limits>
#include <algorithm>
#include <stx/btree_map>
#include <boost/intrusive/list.hpp>
#include <boost/intrusive/list_hook.hpp>

#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>

#include "protocol.h"
#include "def.h"
#include "LinearAllocator.h"
#include "aggr.h"
#include "Reader.h"
#include "util.h"
#include "PageStore.h"

// #define VERIFY_FUTURE_DATA_ACCESS
// #define SINGLE_THREAD

// a set of tunable parameters
enum { N_THREAD = 4 }; // number of worker threads besides main
enum { FORGET_THRESHOLD = 50000000 }; // how many records to "forget" at a time
enum { MAX_SEL_SAMPLES = 20 }; // how many selectivity sample to compute, ~5k records
enum { IDX_SEL_THRESHOLD = 15 }; // column selectivity used to choose an index

// 64-bit hash function
inline uint32_t crc(uint64_t v)
{
	return _mm_crc32_u32(_mm_crc32_u32(0, v & 0xffffffffu), v >> 32);
}

// data log

typedef uint32_t log_ptr_t;

// minimum byte alignment of records in the data log (log2)
enum : unsigned
{
	LOG_ALIGN_BITS = 3
};

// specializations of data structures
typedef LinearAllocator<21> DataLog;
typedef LinearAllocator<20> HeapLog;
typedef LinearAllocator<21> CmdLog;

log_ptr_t pack_record_ptr(size_t offset)
{
	static_assert(sizeof(log_ptr_t) == 4, "");
	assert(alignOffsetBits(offset, LOG_ALIGN_BITS) == 0);
	offset >>= LOG_ALIGN_BITS;
	log_ptr_t rec_ptr = offset;
	assert(rec_ptr == offset);
	return rec_ptr;
}

size_t unpack_record_ptr(log_ptr_t rec_ptr)
{
	return size_t(rec_ptr) << LOG_ALIGN_BITS;
}

// data aggregation

template <AggrOp Op>
struct Aggr: public AggrInt<Op> {};

enum
{
	AGGR_BLOCKING_BITS = 1,
	AGGREGATE_BLOCK = 256 // tunable parameter: aggregation block size
};

struct log_aggr_t
{
	txn_id_t min_txn;
	txn_id_t max_txn;
	value_t minmax[0]; // variable size, min followed by max
	static size_t size(unsigned columns)
	{
		return 0
				+ sizeof(min_txn) + sizeof(max_txn)
				+ alignUpBits(columns, AGGR_BLOCKING_BITS) * sizeof(value_t) * 2;
	}
	value_t *min(unsigned)
	{
		return minmax;
	}
	value_t *max(unsigned columns)
	{
		return minmax + alignUpBits(columns, AGGR_BLOCKING_BITS);
	}
};

// predicate verification building blocks

bool verifyRecord(const Query::Column columns[], unsigned column_count, const value_t *record)
{
	for (auto c = columns, cLimit = c + column_count; c != cLimit; ++c)
	{
		uint64_t tupleValue = record[c->column], queryValue = c->value;
		bool result = true;
		switch (c->op)
		{
		case Query::Column::Equal:
			result = tupleValue == queryValue;
			break;
		case Query::Column::NotEqual:
			result = tupleValue != queryValue;
			break;
		case Query::Column::Less:
			result = tupleValue < queryValue;
			break;
		case Query::Column::LessOrEqual:
			result = tupleValue <= queryValue;
			break;
		case Query::Column::Greater:
			result = tupleValue > queryValue;
			break;
		case Query::Column::GreaterOrEqual:
			result = tupleValue >= queryValue;
			break;
		}
		if (!result)
			return false;
	}
	return true;
}

bool verifyMinMaxRecord(const Query::Column columns[], unsigned col_count, const value_t *min, const value_t *max)
{
	for (auto c = columns, cLimit = c + col_count; c != cLimit; ++c)
	{
		uint64_t queryValue = c->value;
		bool result = true;
		switch (c->op)
		{
		case Query::Column::Equal:
			result = min[c->column] <= queryValue && max[c->column] >= queryValue;
			break;
		case Query::Column::Less:
			result = min[c->column] < queryValue;
			break;
		case Query::Column::LessOrEqual:
			result = min[c->column] <= queryValue;
			break;
		case Query::Column::Greater:
			result = max[c->column] > queryValue;
			break;
		case Query::Column::GreaterOrEqual:
			result = max[c->column] >= queryValue;
			break;
		default:
			break;
		}
		if (!result)
			return false;
	}
	return true;
}

// transaction record
struct log_record_t
{
	enum : uint64_t { DELTXN_DEL = 0, DELTXN_NOTYET = 0xfffffffffffffffful };

	txn_id_t txn; // id of the ransaction that created this record
	txn_id_t del_txn; // id of transaction that deleted that record (for INSERT transactions only)
	char v[]; // data and link fields, used by ugly accessor below
	log_ptr_t *link() { return reinterpret_cast<log_ptr_t *>(&v); }
	const log_ptr_t *link() const { return reinterpret_cast<const log_ptr_t *>(&v); }
	value_t *data(unsigned col_blk) { return reinterpret_cast<value_t *>(link() + col_blk); }
	const value_t *data(unsigned col_blk) const { return reinterpret_cast<const value_t *>(link() + col_blk); }
	int is_delete() const { return del_txn  == DELTXN_DEL; }

	static unsigned size(unsigned columns_aligned)
	{
		return sizeof(txn_id_t) * 2 + alignUpBits(columns_aligned, 1) * sizeof(log_ptr_t) + alignUpBits(columns_aligned, 1) * sizeof(value_t);
	}
};

// defines basic constants for a relation
struct RelationMetadata
{
	unsigned rel_id;
	uint16_t columns;
	uint16_t columns_aligned;
	unsigned tuple_sz;
	unsigned log_rec_sz;

	RelationMetadata(unsigned rel_id, uint16_t columns):
		rel_id(rel_id),
		columns(columns),
		columns_aligned(alignUpBits(columns, 1)),
		tuple_sz(sizeof(value_t) * columns_aligned),
		log_rec_sz(log_record_t::size(columns_aligned)) {}
};


enum { QRY_COL_BLK_BITS = 1 };

// Stage1 to Stage2 messaging protocol

// a tag to distinguish between transaction and validation message, overlaps with the message
struct CmdBase
{
	enum : uint64_t
	{
		TAG_INSERT = -1ul,
		TAG_DELETE = -2ul
	};

	uint64_t tag;
	bool is_txn() const { return tag >= TAG_DELETE; }
};

// transaction message
struct TxnCmd
{
	uint64_t tag; // -1
	txn_id_t txn;
	const uint64_t *values;
	uint32_t row_cnt;
	uint32_t pad;
};

static_assert(sizeof(TxnCmd) == 32, "");

// validation message for a query
struct ValHdr
{
	uint64_t from, to; // transaction boundaries
	uint32_t val_index; // index of the validation in the current batch used to return the result
	int16_t index_col; // column id of the used index
	uint16_t aggr_sz; // current position in the aggregation block
	union 
	{ 
		uint64_t data_limit; // upper data bound for scan
		uint32_t index_probe_ptr; // starting pointer for index lookup
	};
	const Query *qry; // pointer to the original query
	uint32_t col_count() const { return qry->columnCount; }
	const Query::Column *columns() const { return qry->columns; }
	static unsigned size(unsigned) { return sizeof(ValHdr); }
	void init_col(const Query &qry)
	{
		this->qry = &qry;
	}
};

static_assert(sizeof(ValHdr) == 40, "");

// complete query validation using an index
static
bool verify_index(const RelationMetadata& meta, const DataLog &data_log, const ValHdr &qry)
{
	size_t offset = unpack_record_ptr(qry.index_probe_ptr);
	while (data_log.valid(offset))
	{
		const auto &log_record = *(log_record_t *)data_log.resolve(offset);
		if (log_record.txn <= qry.to)
		{
			if (log_record.txn < qry.from)
				break;
			auto next_ptr = log_record.link()[qry.index_col];
			auto next_ofs = unpack_record_ptr(next_ptr);
			if (likely(data_log.valid(next_ofs)))
			{
				const auto &next_rec = *(log_record_t *)data_log.resolve(next_ofs);
				_mm_prefetch(&next_rec, _MM_HINT_NTA);
			}
			if (verifyRecord(qry.columns(), qry.col_count(), log_record.data(meta.columns_aligned)))
				goto queryMatch;
		}
		offset = unpack_record_ptr(log_record.link()[qry.index_col]);
	}
	return false;

queryMatch:
	return true;
}

// complete query validation, using table scan
static
bool verify_scan(const RelationMetadata& meta, const DataLog &data_log, size_t data_start, const ValHdr &qry)
{
	size_t offset = qry.data_limit;
	unsigned aggr_sz = qry.aggr_sz;
	bool skip_block = false;

	if (aggr_sz == 0)
		goto endOfBlock;

	while (1)
	{
		assert(aggr_sz > 0);
		if (skip_block)
		{
			offset = data_log.move_back(offset, meta.log_rec_sz, aggr_sz);
			auto &log_record = *(log_record_t *) data_log.resolve(offset);
			if (log_record.txn < qry.from)
				goto queryMismatch;
		}
		else
		{
			do
			{
				offset = data_log.move_back(offset, meta.log_rec_sz);
				auto &log_record = *(log_record_t *) data_log.resolve(offset);
				if (log_record.txn <= qry.to)
				{
					if (log_record.txn < qry.from)
						goto queryMismatch;
					if (verifyRecord(qry.columns(), qry.col_count(), log_record.data(meta.columns_aligned)))
						goto queryMatch;
				}
			} while (--aggr_sz);
		}
endOfBlock:
		if (offset == data_start)
			break;
		offset = data_log.move_back(offset, log_aggr_t::size(meta.columns_aligned));
		auto &aggr_record = *(const log_aggr_t *)data_log.resolve(offset);
		skip_block = aggr_record.min_txn > qry.to
				|| !verifyMinMaxRecord(qry.columns(), qry.col_count(), aggr_record.minmax, aggr_record.minmax + meta.columns_aligned);
		aggr_sz = AGGREGATE_BLOCK;
	}

queryMismatch:
	return false;

queryMatch:
	return true;
}

// used by 1st thread exclusively
class RelationStage1
{
public:

	RelationStage1(const RelationMetadata &meta, CmdLog &cmd_log);

	void prepare_delete(Reader &in, txn_id_t txn, unsigned row_count);
	void prepare_insert(Reader &in, txn_id_t txn, unsigned row_count);
	void prepare_query(unsigned val_index, txn_id_t from, txn_id_t to, const Query &q);
	void cleanup(size_t cmd_processed);
	void schedule_stage2_if_work_pending();

private:

	const RelationMetadata m_meta;
	CmdLog &m_cmd_log;
	size_t m_scheduled_val_pos;
	txn_id_t m_next_val_txn;
	long pad[1];

	enum { MAX_TXN_BLOCK = Reader::READ_SZ / sizeof(value_t) };

	void schedule_stage2_lazily_unlocked();
};

RelationStage1::RelationStage1(const RelationMetadata &meta, CmdLog &cmd_log):
	m_meta(meta),
	m_cmd_log(cmd_log),
	m_scheduled_val_pos(cmd_log.head()),
	m_next_val_txn(0)
{
}

// process an delete by the main thread
void RelationStage1::prepare_delete(Reader &in, txn_id_t txn, unsigned row_count)
{
#ifdef VERIFY_FUTURE_DATA_ACCESS
	if (unlikely(txn < rel.next_val_txn))
	{
		// this DOES happen in some datasets
		fprintf(stderr, "Transaction %lu was referenced by previous validations\n", t.transactionId);
		abort();
	}
#endif
	unsigned block = row_count % MAX_TXN_BLOCK;
	for (unsigned c = 0; c != row_count; c += block, block = MAX_TXN_BLOCK)
	{
		auto *keys = in.read<value_t>(sizeof(value_t) * block);
		auto ptr = m_cmd_log.allocate(sizeof(TxnCmd));
		auto &cmd = *(TxnCmd *) m_cmd_log.resolve(ptr);
		cmd.tag = CmdBase::TAG_DELETE;
		cmd.txn = txn;
		cmd.row_cnt = block;
		cmd.values = keys;
	}
	schedule_stage2_lazily_unlocked();
}

// process an insert by the main thread
void RelationStage1::prepare_insert(Reader &in, txn_id_t txn, unsigned row_count)
{
#ifdef VERIFY_FUTURE_DATA_ACCESS
	if (unlikely(t.transactionId < rel.next_val_txn))
	{
		fprintf(stderr, "Transaction %lu was referenced by previous validations\n", t.transactionId);
		abort();
	}
#endif
	unsigned rows_blk = MAX_TXN_BLOCK / m_meta.columns;
	unsigned block = row_count % rows_blk;
	for (unsigned c = 0; c != row_count; c += block, block = rows_blk)
	{
		auto *data = in.read<value_t>(sizeof(value_t) * block * m_meta.columns);
		auto ptr = m_cmd_log.allocate(sizeof(TxnCmd));
		auto &cmd = *(TxnCmd *) m_cmd_log.resolve(ptr);
		cmd.tag = CmdBase::TAG_INSERT;
		cmd.txn = txn;
		cmd.row_cnt = block;
		cmd.values = data;
	}
	schedule_stage2_lazily_unlocked();
}

// process an validation query by the main thread
void RelationStage1::prepare_query(unsigned val_index, txn_id_t from, txn_id_t to, const Query &q)
{
	auto &hdr = *(ValHdr*) m_cmd_log.resolve(m_cmd_log.allocate(ValHdr::size(q.columnCount)));
	hdr.val_index = val_index;
	hdr.from = from;
	hdr.to = to;
	hdr.init_col(q);
#ifdef VERIFY_FUTURE_DATA_ACCESS
	if (v.to >= m_next_val_txn)
		m_next_val_txn = to + 1;
#endif
	schedule_stage2_lazily_unlocked();
}

class RelationStage2;

class Stage2VerifyTask: public boost::intrusive::list_base_hook<>
{
public:

	Stage2VerifyTask(RelationStage2 &rel, size_t val_limit, txn_id_t min_valid_txn):
		m_rel(&rel), m_val_limit(val_limit), m_min_valid_txn(min_valid_txn) {}

	void reinit(RelationStage2 &rel, size_t val_limit, txn_id_t min_valid_txn)
	{
		m_rel = &rel;
		m_val_limit = val_limit;
		m_min_valid_txn = min_valid_txn;
		m_satisfied.clear();
	}
	void merge(std::vector<char> &val) const
	{
		for (auto &i : m_satisfied)
		{
			assert(i < val.size());
			val[i] = '1';
		}
	}
	void run();

private:

	RelationStage2 *m_rel;
	size_t m_val_limit;
	txn_id_t m_min_valid_txn;
	std::vector<unsigned> m_satisfied;
};

typedef boost::intrusive::list<Stage2VerifyTask> Stage2VerifyTaskList;


struct RelationStage2
{
	boost::intrusive::list_member_hook<> stage2_sched_hook;
	Stage2VerifyTaskList stage2_tasks;
	bool stage2_busy;

	CmdLog cmd_log; // incoming messages from the main thread

	const RelationMetadata meta;
	DataLog data_log; // transaction data log
	HeapLog heap; // scratch allocator for "forgotten" but not deleted records
	HeapLog::offset_t heap_free_list; // free list for the allocator above
	enum : HeapLog::offset_t { HEAP_EMPTY = 1 }; // this is not a valid offset value because real offsets are 8-aligned

	size_t val_processed; // monotonic valid data upper bound
	size_t data_forgotten; // monotonic valid data lower bound

	unsigned aggr_cnt2; // current position in the aggregation block
	log_aggr_t *aggr_rec; // current aggregation record

	typedef stx::btree_map<value_t, size_t> PrimaryKey;
	PrimaryKey primary_key; // keys of existing records are stored here

	typedef std::vector<log_ptr_t> hash_table_t;
	std::vector<hash_table_t> hash; // hash table for each column

	struct ColStat
	{
		unsigned sel_smooth; // smoothed selectivity, normalized by number of insertions
		unsigned sel_current; // current selectivity
		unsigned hash_bits; // current hash table size (log2)
	};

	std::vector<ColStat> col_stat;
	unsigned sel_interval_cnt;
	unsigned sel_interval_max;
	unsigned sel_samples;

	uint64_t padding[2];

	RelationStage2(const RelationMetadata &meta, PageStore &page_store, unsigned space_bits);

	void prepare_table(size_t val_limit, txn_id_t min_valid_txn, std::vector<unsigned> &satisfied_validations);
	void forget(txn_id_t min_valid_txn);
	void schedule(size_t val_limit, txn_id_t min_valid_txn, std::vector<unsigned> &satisfied_validations);

private:

	HeapLog::offset_t heap_alloc();
	void heap_free(HeapLog::offset_t);
	log_record_t &insert_log_record(txn_id_t txn, txn_id_t del_txn, size_t &ofs);
	void process_record(log_record_t &rec, log_ptr_t rec_ptr);
	void reset_aggregates();
	void start_sel_interval();
	void prepare_query(size_t data_limit, ValHdr &hdr);
};

RelationStage2::RelationStage2(const RelationMetadata &meta, PageStore &page_store, unsigned space_bits):
	stage2_busy(false),
	cmd_log(space_bits, &page_store),
	meta(meta),
	data_log(space_bits, &page_store),
	heap(space_bits, &page_store),
	heap_free_list(HEAP_EMPTY),
	val_processed(cmd_log.head()),
	data_forgotten(data_log.head()),
	aggr_cnt2(0),
	aggr_rec((log_aggr_t *) malloc(log_aggr_t::size(meta.columns_aligned))),
	hash(meta.columns),
	col_stat(meta.columns),
	sel_interval_cnt(0),
	sel_interval_max(1),
	sel_samples(0)
{
	aggr_rec->min_txn = 0;
	aggr_rec->max_txn = 0;
	reset_aggregates();
	for (auto &i : col_stat)
	{
		i.hash_bits = 4; // starting hash table size (log2) - 16 slots
		i.sel_smooth = 0;
		i.sel_current = 0;
	}
	for (auto &i : hash)
	{
		i.resize(size_t(1) << 4);
		std::fill(i.begin(), i.end(), pack_record_ptr(data_log.move_back(data_log.block_tail(), 1 << LOG_ALIGN_BITS)));
	}
}

// allocate a temporary record
HeapLog::offset_t RelationStage2::heap_alloc()
{
	HeapLog::offset_t ofs;
	if (heap_free_list == HEAP_EMPTY)
		ofs = heap.allocate(meta.tuple_sz);
	else
	{
		ofs = heap_free_list;
		heap_free_list = *(HeapLog::offset_t *)heap.resolve(heap_free_list);
	}
	return ofs;
}

// free the temeporary record
void RelationStage2::heap_free(HeapLog::offset_t ofs)
{
	void *p = heap.resolve(ofs);
	*(HeapLog::offset_t *)p = heap_free_list;
	heap_free_list = ofs;
}

// add a new record to the transaction list, updating stats as appropriate
log_record_t &RelationStage2::insert_log_record(txn_id_t txn, txn_id_t del_txn, size_t &rec_ofs)
{
	// put aggregate record first to ensure there is always at least one record after it
	if (unlikely(aggr_cnt2 == AGGREGATE_BLOCK))
	{
		size_t aggr_ofs = data_log.allocate(log_aggr_t::size(meta.columns_aligned)); // NOTE: not aligned for SIMD
		auto &aggr_record = *(log_aggr_t *)data_log.resolve(aggr_ofs);
		aggr_record.min_txn = aggr_rec->min_txn;
		aggr_record.max_txn = aggr_rec->max_txn;
		copyArray<1>(aggr_record.minmax, aggr_rec->minmax, meta.columns_aligned * 2); // NOTE: 8bt alignment
		reset_aggregates();
		if (sel_samples < MAX_SEL_SAMPLES && ++sel_interval_cnt >= sel_interval_max)
			start_sel_interval();
		aggr_cnt2 = 0;
		aggr_rec->min_txn = txn;
	}
	else
	{
		if (aggr_cnt2 == 0)
			aggr_rec->min_txn = txn;
	}
	rec_ofs = data_log.allocate(meta.log_rec_sz);
	auto &record = *(log_record_t *) data_log.resolve(rec_ofs);
	record.txn = txn;
	record.del_txn = del_txn;
	++aggr_cnt2;
	return record;
}

// update the hash table and aggregate record on a new record
void RelationStage2::process_record(log_record_t &rec, log_ptr_t rec_ptr)
{
	const auto *data = rec.data(meta.columns_aligned);
	auto *link = rec.link();
	for (unsigned i = 0; i != meta.columns; ++i)
	{
		if (sel_samples >= MAX_SEL_SAMPLES && col_stat[i].sel_smooth < IDX_SEL_THRESHOLD)
			continue;
		auto h = crc(data[i]);
		auto &s = col_stat[i];
		auto offset = alignOffsetBits(h, s.hash_bits);
		auto &hash_slot = hash[i][offset];
		s.sel_current += (rec_ptr - hash_slot) & (data_log.ring_mask() >> LOG_ALIGN_BITS);
		link[i] = hash_slot;
		hash_slot = rec_ptr;
	}
	aggr_rec->max_txn = rec.txn;
	::aggregate<Aggr<MIN>, AGGR_BLOCKING_BITS>(meta.columns_aligned, aggr_rec->min(meta.columns_aligned), data);
	::aggregate<Aggr<MAX>, AGGR_BLOCKING_BITS>(meta.columns_aligned, aggr_rec->max(meta.columns_aligned), data);
}

void RelationStage2::reset_aggregates()
{
	for (unsigned i = 0; i != meta.columns_aligned; ++i)
	{
		aggr_rec->min(meta.columns_aligned)[i] = Aggr<MIN>::init();
		aggr_rec->max(meta.columns_aligned)[i] = Aggr<MAX>::init();
	}
}
void RelationStage2::start_sel_interval()
{
	for (unsigned i = 0; i != meta.columns; i++)
	{
		auto &s = col_stat[i];
		s.sel_smooth -= s.sel_smooth / 4;
		s.sel_smooth += s.sel_current / 4 / (meta.columns * sel_interval_cnt * AGGREGATE_BLOCK);
		s.sel_current = 0;
		unsigned current_size = 1u << s.hash_bits;
		if (s.sel_smooth > current_size - current_size / 2 && s.hash_bits < 12)
		{
			auto &h = hash[i];
			h.resize(current_size * 2);
			for (unsigned i = 0; i != current_size; ++i)
				h[i + current_size] = h[i];
			s.hash_bits++;
		}
	}
	sel_interval_max = 4;
	sel_interval_cnt = 0;
	sel_samples++;
}

inline
uint64_t op_sel(unsigned op, uint64_t x)
{
	uint64_t zero = 0;
	asm("cmpl %1,%2 ; cmovne %3,%0"
		:"=r" (x)
		:"i" (Query::Column::Equal), "r" (op), "r" (zero), "0" (x));
	return x;
}

// choose whether to use an index (and which) for a query and do the hash lookup, if so
void RelationStage2::prepare_query(size_t data_limit, ValHdr &qry)
{
	if (likely(qry.col_count() != 0))
	{
		uint64_t hack = 0;
		for (unsigned i = 0; i != qry.col_count(); i++)
		{
			uint64_t x = op_sel(qry.columns()[i].op, (uint64_t(col_stat[qry.columns()[i].column].sel_smooth) << 32) | i);
			if (x > hack)
				hack = x;
		}
		if (hack >= uint64_t(IDX_SEL_THRESHOLD) << 32)
		{
			// index has a good enough selectivity, use it
			unsigned index_col = qry.columns()[hack & 0xffffffffu].column;
			value_t index_value = qry.columns()[hack & 0xffffffffu].value;
			uint32_t h = crc(index_value);
			uint32_t offset = alignOffsetBits(h, col_stat[index_col].hash_bits);
			auto rec_ptr = hash[index_col][offset];
			qry.index_col = index_col;
			qry.index_probe_ptr = rec_ptr;
			return;
		}
	}

	// no appropriate index, do a scan
	qry.index_col = -1;

	// verify min/max immediately because the last incomplete block will change before queries can be validated
	if (aggr_rec->min_txn > qry.to
			|| !verifyMinMaxRecord(qry.columns(), qry.col_count(), aggr_rec->min(meta.columns_aligned), aggr_rec->max(meta.columns_aligned)))
	{
		qry.data_limit = data_log.move_back(data_limit, meta.log_rec_sz, aggr_cnt2);
		qry.aggr_sz = 0;
	}
	else
	{
		qry.data_limit = data_limit;
		qry.aggr_sz = aggr_cnt2;
	}
}

void release_tasks_unlocked(RelationStage2 &rel);

// process a batch of messages from the main thread
void RelationStage2::prepare_table(size_t val_limit, txn_id_t min_valid_txn, std::vector<unsigned> &satisfied_validations)
{
	for (size_t vp = val_processed; vp != val_limit; )
	{
		auto &cmd = *(CmdBase*) cmd_log.resolve(vp);
		if (unlikely(cmd.is_txn()))
		{
			auto &txn = reinterpret_cast<const TxnCmd &>(cmd);
			if (cmd.tag == CmdBase::TAG_INSERT)
			{
				for (auto p = txn.values; p != txn.values + txn.row_cnt * meta.columns; p += meta.columns)
				{
					size_t rec_ofs;
					if (txn.txn >= min_valid_txn)
					{
						auto &record = insert_log_record(txn.txn, log_record_t::DELTXN_NOTYET, rec_ofs);
						memcpy(record.data(meta.columns_aligned), p, sizeof(value_t) * meta.columns);
						process_record(record, pack_record_ptr(rec_ofs));
					}
					else
					{
						rec_ofs = heap_alloc();
						memcpy((value_t *)heap.resolve(rec_ofs), p, sizeof(value_t) * meta.columns);
						rec_ofs |= 0x8000000000000000ul;
					}
					bool inserted = primary_key.insert2(p[0], rec_ofs).second;
					assert(inserted);
					(void) inserted;
				}
			}
			else
			{
				for (auto p = txn.values; p != txn.values + txn.row_cnt; ++p)
				{
					auto it = primary_key.find(*p);
					if (it != primary_key.end())
					{
						if (txn.txn >= min_valid_txn)
						{
							size_t rec_ofs;
							auto &record = insert_log_record(txn.txn, log_record_t::DELTXN_DEL, rec_ofs);
							auto org_ofs = it->second;
							if (unlikely(org_ofs & 0x8000000000000000ul))
							{
								org_ofs &= 0x7ffffffffffffffful;
								auto *org_values = (value_t *)heap.resolve(org_ofs);
								copyArray<1>(record.data(meta.columns_aligned), org_values, meta.columns_aligned);
								heap_free(org_ofs);
							}
							else
							{
								auto &org_rec = *(log_record_t *) data_log.resolve(org_ofs);
								assert(org_rec.del_txn == log_record_t::DELTXN_NOTYET);
								org_rec.del_txn = txn.txn;
								copyArray<1>(record.data(meta.columns_aligned), org_rec.data(meta.columns_aligned), meta.columns_aligned);
							}
							process_record(record, pack_record_ptr(rec_ofs));
						}
						primary_key.erase(it);
					}
				}
			}
			vp = cmd_log.move(vp, sizeof(TxnCmd));
		}
		else
		{
			auto &qry = reinterpret_cast<ValHdr &>(cmd);
			prepare_query(data_log.head(), qry);
			vp = cmd_log.move(vp, sizeof(ValHdr));
		}
	}

	if (data_log.diff_gross(data_log.head(), data_forgotten) >= FORGET_THRESHOLD)
		forget(min_valid_txn);

	auto saved_start = val_processed;

	// some data may be left unprocessed if not referenced by the last val
	val_processed = val_limit;

	// schedule Stage3 task
	release_tasks_unlocked(*this);

	for (size_t vp = saved_start; vp != val_limit; )
	{
		auto &cmd = *(CmdBase*) cmd_log.resolve(vp);
		if (likely(!cmd.is_txn()))
		{
			auto &qry = reinterpret_cast<const ValHdr &>(cmd);
			bool r = qry.index_col != -1 ? verify_index(meta, data_log, qry)
					: verify_scan(meta, data_log, data_forgotten, qry);
			if (r)
				satisfied_validations.push_back(qry.val_index);
			vp = cmd_log.move(vp, sizeof(ValHdr));
		}
		else
		{
			vp = cmd_log.move(vp, sizeof(TxnCmd));
		}
	}
}

// discard processed messages
void RelationStage1::cleanup(size_t cmd_processed)
{
	m_cmd_log.collect(cmd_processed);
}

// discard data that has been "forgotten"
void RelationStage2::forget(txn_id_t min_valid_txn)
{
	size_t end = data_log.move_back(data_log.head(), meta.log_rec_sz, aggr_cnt2);
	size_t dp = data_forgotten;
	while (dp != end
			&& data_log.diff_gross(end, dp) >= FORGET_THRESHOLD)
	{
		size_t aggr_ptr = data_log.move(dp, meta.log_rec_sz, AGGREGATE_BLOCK);
		auto &aggr_record = *(const log_aggr_t *)data_log.resolve(aggr_ptr);
		if (aggr_record.max_txn >= min_valid_txn)
			break;
		while (dp != aggr_ptr)
		{
			auto &record = *(const log_record_t *) data_log.resolve(dp);
			assert(record.txn >= aggr_record.min_txn);
			assert(record.txn <= aggr_record.max_txn);
			if (record.del_txn == log_record_t::DELTXN_NOTYET)
			{
				auto data = record.data(meta.columns_aligned);
				auto heap_ofs = heap_alloc();
				copyArray<1>((value_t *)heap.resolve(heap_ofs), data, meta.columns_aligned);
				auto it = primary_key.lower_bound(data[0]);
				assert(it != primary_key.end());
				assert(it->first == data[0]);
				assert(it->second == dp);
				it.data() = heap_ofs | 0x8000000000000000ul;
			}
			dp = data_log.move(dp, meta.log_rec_sz);
		}
		dp = data_log.move(aggr_ptr, log_aggr_t::size(meta.columns_aligned));
	}
	data_log.collect(dp);
	auto invalid_ptr = pack_record_ptr(data_log.move_back(data_log.block_tail(), 1 << LOG_ALIGN_BITS));
	for (unsigned i = 0; i != meta.columns; i++)
	{
		for (auto &h : hash[i])
		{
			if (!data_log.valid(unpack_record_ptr(h)))
				h = invalid_ptr;
		}
	}
	data_forgotten = dp;
}

void Stage2VerifyTask::run()
{
	m_rel->prepare_table(m_val_limit, 0, m_satisfied);
}

// all the state goes here
std::vector<RelationMetadata> rel_meta;
std::vector<RelationStage1> rel_stage1;
std::vector<RelationStage2> rel_stage2;

typedef boost::intrusive::list<RelationStage2,
	boost::intrusive::member_hook<RelationStage2, boost::intrusive::list_member_hook<>, &RelationStage2::stage2_sched_hook>> Stage2RelationList;
Stage2RelationList stage2_rel_pending;

// return queue - access must be synchronized
Stage2VerifyTaskList return_tasks;

// the following belongs to the main thread
txn_id_t min_valid_txn = 0;
unsigned tasks_running = 0;
Stage2VerifyTaskList free_tasks;
uint64_t start_val = 0; // for debugging, assumes validation id are consecutive and start at 0
std::vector<char> results;
std::vector<uint64_t> validations;
txn_id_t lastTransaction = 0;
uint64_t lastValidation = 0;

pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t res_cond = PTHREAD_COND_INITIALIZER;

extern PageStore g_page_store;

struct Thread: public boost::intrusive::list_base_hook<>
{
	unsigned i;
	pthread_t thr;
	pthread_cond_t cond;
	RelationStage2 *work;
};

typedef boost::intrusive::list<Thread> ThreadList;
ThreadList idle_threads;

void *thread_func(void *arg)
{
	Thread &thread = *(Thread*)arg;
	pthread_mutex_lock(&m);
	while (1)
	{
		while (thread.work == NULL)
		{
			assert(thread.is_linked());
			assert(!idle_threads.empty());
			pthread_cond_wait(&thread.cond, &m);
		}
		RelationStage2 &rel = *thread.work;
		assert(!rel.stage2_tasks.empty());
		assert(!rel.stage2_sched_hook.is_linked());
		Stage2VerifyTask &task = rel.stage2_tasks.front();
		pthread_mutex_unlock(&m);
		task.run();
		pthread_mutex_lock(&m);
		if (return_tasks.empty())
			pthread_cond_signal(&res_cond);
		return_tasks.push_front(task);
		if (!stage2_rel_pending.empty())
		{
			thread.work = &stage2_rel_pending.front();
			stage2_rel_pending.pop_front();
		}
		else
		{
			thread.work = NULL;
			idle_threads.push_front(thread);
		}
	}
	pthread_mutex_unlock(&m);
	return 0;
}

Thread thread[N_THREAD];

static void processDefineSchema(const DefineSchema& d)
{
	for (unsigned i = 0; i != N_THREAD; ++i)
	{
		auto &t = thread[i];
		t.i = i;
		t.work = NULL;
		pthread_cond_init(&t.cond, NULL);
		idle_threads.push_back(t);
		pthread_create(&t.thr, NULL, thread_func, &t);
	}

	rel_meta.reserve(d.relationCount);
	rel_stage1.reserve(d.relationCount);
	rel_stage2.reserve(d.relationCount);

	unsigned rel_log2 = d.relationCount <= 1 ? 0 : 32 - __builtin_clz(d.relationCount - 1);
	unsigned space_bits = std::min(35u, 46u - rel_log2 - 3u);

	for (uint32_t i = 0; i != d.relationCount; ++i)
	{
		rel_meta.emplace_back(i, d.columnCounts[i]);
		rel_stage2.emplace_back(rel_meta.back(), g_page_store, space_bits);
		rel_stage1.emplace_back(rel_meta.back(), rel_stage2.back().cmd_log);
	}
	results.reserve(16384);
	validations.reserve(16384);
}

void prepareTransaction(Reader &in, const Transaction& t)
{
	for (uint32_t index = 0; index != t.deleteCount; ++index)
	{
		auto o = *in.read<TransactionOperationDelete>();
		auto &rel = rel_stage1[o.relationId];
		rel.prepare_delete(in, t.transactionId, o.rowCount);
	}
	for (uint32_t index = 0; index != t.insertCount; ++index)
	{
		auto o = *in.read<TransactionOperationInsert>();
		auto &rel = rel_stage1[o.relationId];
		rel.prepare_insert(in, t.transactionId, o.rowCount);
	}
}

void scheduleRel(RelationStage2 &rel)
{
	assert(!rel.stage2_sched_hook.is_linked());
	if (!idle_threads.empty())
	{
		auto &thread = idle_threads.front();
		assert(thread.work == NULL);
		assert(stage2_rel_pending.empty());
		thread.work = &rel;
		pthread_cond_signal(&thread.cond);
		idle_threads.pop_front();
	}
	else
	{
		stage2_rel_pending.push_front(rel);
	}
}

void schedule_stage2(RelationStage2 &rel, size_t cmd_limit)
{
#ifndef SINGLE_THREAD
	tasks_running++;
	Stage2VerifyTask *task;
	if (!free_tasks.empty())
	{
		task = &free_tasks.front();
		free_tasks.pop_front();
		task->reinit(rel, cmd_limit, min_valid_txn);
	}
	else
	{
		task = new Stage2VerifyTask(rel, cmd_limit, min_valid_txn);
	}
	if (rel.stage2_tasks.empty())
		scheduleRel(rel);
	rel.stage2_tasks.push_back(*task);
#else
	pthread_mutex_unlock(&m);
	rel.prepare_table(val_limit);
	pthread_mutex_lock(&m);
#endif
}

// schedule a Stage3 task (final verfication)
void release_tasks_unlocked(RelationStage2 &rel)
{
#ifndef SINGLE_THREAD
	pthread_mutex_lock(&m);
	assert(!rel.stage2_tasks.empty());
	assert(!rel.stage2_sched_hook.is_linked());
	rel.stage2_tasks.pop_front();
	if (!rel.stage2_tasks.empty())
		scheduleRel(rel);
	pthread_mutex_unlock(&m);
#endif
}

// schedule a task in the Stage2 is there is pending data above a reasonable size
void RelationStage1::schedule_stage2_lazily_unlocked()
{
	if (m_cmd_log.diff_gross(m_cmd_log.head(), m_scheduled_val_pos) >= 20000)
	{
		pthread_mutex_lock(&m);
		schedule_stage2(rel_stage2[m_meta.rel_id], m_cmd_log.head());
		pthread_mutex_unlock(&m);
		m_scheduled_val_pos = m_cmd_log.head();
	}
}

// schedule a task in the Stage2 is there is any pending data
void RelationStage1::schedule_stage2_if_work_pending()
{
	if (m_cmd_log.head() != m_scheduled_val_pos)
	{
		schedule_stage2(rel_stage2[m_meta.rel_id], m_cmd_log.head());
		m_scheduled_val_pos = m_cmd_log.head();
	}
}

void prepareValidationQueries(const ValidationQueries& v)
{
	unsigned val_index = results.size();
	validations.push_back(v.validationId);
	results.push_back('0');
	const char* qp = v.queries;
	for (unsigned i = 0; i != v.queryCount; ++i)
	{
		const Query &q = *reinterpret_cast<const Query *>(qp); // NOTE: unaligned
		auto &rel = rel_stage1[q.relationId];
		rel.prepare_query(val_index, v.from, v.to, q);
		qp += sizeof(Query) + sizeof(Query::Column) * q.columnCount;
	}
}

static
void processForget(txn_id_t txn)
{
	min_valid_txn = txn;
}

void processInput(Reader &in, uint64_t len, std::queue<char> *result = NULL)
{
	uint64_t start = in.count();
	while (in.count() < start + len)
	{
		auto head = *in.read<MessageHead>();
		uint64_t next = in.count() + head.messageLen;
		switch (head.type)
		{
		case MessageHead::Done:
			goto fin;
		case MessageHead::DefineSchema:
		{
			const auto& t = *in.read<DefineSchema>(head.messageLen);
			processDefineSchema(t);
			break;
		}
		case MessageHead::Transaction:
		{
			auto t = *in.read<Transaction>();
			prepareTransaction(in, t);
			break;
		}
		case MessageHead::ValidationQueries:
		{
			const auto &v = *in.read<ValidationQueries>(head.messageLen);
			prepareValidationQueries(v);
			break;
		}
		case MessageHead::Flush:
		{
			pthread_mutex_lock(&m);
			for (auto &i : rel_stage1)
				i.schedule_stage2_if_work_pending();
			pthread_mutex_unlock(&m);
			auto f = *in.read<Flush>();
			auto bookmark = in.bookmark();
			unsigned n = results.size();
			while (n != 0 && validations[n - 1] > f.validationId)
				n--;
			while (tasks_running != 0)
			{
				pthread_mutex_lock(&m);
				while (return_tasks.empty())
					pthread_cond_wait(&res_cond, &m);
				Stage2VerifyTaskList completed_tasks;
				completed_tasks.swap(return_tasks);
				pthread_mutex_unlock(&m);
				for (auto &i : completed_tasks)
				{
					assert(tasks_running > 0);
					tasks_running--;
					i.merge(results);
				}
				free_tasks.splice(free_tasks.begin(), completed_tasks);
			}
			for (unsigned i = 0; i != rel_meta.size(); i++)
				rel_stage1[i].cleanup(rel_stage2[i].val_processed);
			while (n != 0)
			{
				int rc = write(STDOUT_FILENO, results.data(), n);
				if (rc < 0 && errno == EINTR)
					continue;
				if (rc <= 0)
					abort();
				if (result != NULL)
					for (int i = 0; i != rc; i++)
						result->push(results[i]);
				results.erase(results.begin(), results.begin() + rc);
				validations.erase(validations.begin(), validations.begin() + rc);
				start_val += rc;
				n -= rc;
			}
			in.release(bookmark);
			break;
		}
		case MessageHead::Forget:
			processForget(in.read<Forget>()->transactionId + 1);
			break;
		default:
			fprintf(stderr, "malformed message" "\n");
			abort();
		}
		assert(next == in.count());
		(void) next;
	}

fin:

	struct timespec end_time;
	clock_gettime(CLOCK_THREAD_CPUTIME_ID, &end_time);
}

void terminate()
{
	idle_threads.clear();
}
